/**
 * Copyright 2013-2015, Facebook, Inc.
 * All rights reserved.
 *
 * This source code is licensed under the BSD-style license found in the
 * LICENSE file in the root directory of this source tree. An additional grant
 * of patent rights can be found in the PATENTS file in the same directory.
 *
 * @typechecks static-only
 * @providesModule ReactServerAsyncRendering
 */
'use strict';

var _createClass = (function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ('value' in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; })();

var _get = function get(_x, _x2, _x3) { var _again = true; _function: while (_again) { var object = _x, property = _x2, receiver = _x3; desc = parent = getter = undefined; _again = false; if (object === null) object = Function.prototype; var desc = Object.getOwnPropertyDescriptor(object, property); if (desc === undefined) { var parent = Object.getPrototypeOf(object); if (parent === null) { return undefined; } else { _x = parent; _x2 = property; _x3 = receiver; _again = true; continue _function; } } else if ('value' in desc) { return desc.value; } else { var getter = desc.get; if (getter === undefined) { return undefined; } return getter.call(receiver); } } };

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError('Cannot call a class as a function'); } }

function _inherits(subClass, superClass) { if (typeof superClass !== 'function' && superClass !== null) { throw new TypeError('Super expression must either be null or a function, not ' + typeof superClass); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, enumerable: false, writable: true, configurable: true } }); if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass; }

var ReactDefaultBatchingStrategy = require('./ReactDefaultBatchingStrategy');
var ReactElement = require('./ReactElement');
var ReactInstanceHandles = require('./ReactInstanceHandles');
var ReactMarkupChecksum = require('./ReactMarkupChecksum');
var ReactServerBatchingStrategy = require('./ReactServerBatchingStrategy');
var ReactServerRenderingTransaction = require('./ReactServerRenderingTransaction');
var ReactUpdates = require('./ReactUpdates');

var emptyObject = require('fbjs/lib/emptyObject');
var instantiateReactComponent = require('./instantiateReactComponent');
var invariant = require('fbjs/lib/invariant');

var rollingAdler32 = require('./rollingAdler32');
var stream = require('stream');

// this is a pass through stream that can calculate the hash that is used to
// checksum react server-rendered elements.

var Adler32Stream = (function (_stream$Transform) {
  _inherits(Adler32Stream, _stream$Transform);

  function Adler32Stream(options) {
    var _this = this;

    _classCallCheck(this, Adler32Stream);

    _get(Object.getPrototypeOf(Adler32Stream.prototype), 'constructor', this).call(this, options);
    this.rollingHash = rollingAdler32("");
    this.on("end", function () {
      _this.done = true;
    });
  }

  _createClass(Adler32Stream, [{
    key: '_transform',
    value: function _transform(chunk, encoding, next) {
      this.rollingHash = rollingAdler32(chunk.toString("utf-8"), this.rollingHash);
      this.push(chunk);
      next();
    }

    // returns a promise of a hash that resolves when this readable piped into this is finished.
  }, {
    key: 'hash',
    get: function () {
      var _this2 = this;

      if (this.done) {
        return Promise.resolve(this.rollingHash.hash());
      }
      return new Promise(function (resolve, reject) {
        _this2.on("end", function () {
          resolve(_this2.rollingHash.hash());
        });
      });
    }
  }]);

  return Adler32Stream;
})(stream.Transform);

var RenderStream = (function (_stream$Readable) {
  _inherits(RenderStream, _stream$Readable);

  function RenderStream(componentInstance, id, transaction, context, options) {
    _classCallCheck(this, RenderStream);

    _get(Object.getPrototypeOf(RenderStream.prototype), 'constructor', this).call(this, options);
    this.buffer = "";
    this.componentInstance = componentInstance;
    this.id = id;
    this.transaction = transaction;
    this.context = context;
  }

  /**
   * @param {ReactElement} element
   * @return {string} the HTML markup
   */

  _createClass(RenderStream, [{
    key: '_read',
    value: function _read(n) {
      var _this3 = this;

      // it's possible that the last chunk added bumped the buffer up to > 2 * n, which means we will
      // need to go through multiple read calls to drain it down to < n.
      if (this.buffer.length >= n) {
        this.push(this.buffer.substring(0, n));
        this.buffer = this.buffer.substring(n);
        return;
      }
      if (!this.continuation) {
        // start the rendering chain.
        this.componentInstance.mountComponentAsync(this.id, this.transaction, this.context, function (text, cb) {
          _this3.buffer += text;
          if (_this3.buffer.length >= n) {
            _this3.push(_this3.buffer.substring(0, n));
            _this3.buffer = _this3.buffer.substring(n);
            _this3.continuation = cb;
          } else {
            // continue rendering until we have enough text to call this.push().
            // TODO: sometimes do this as process.nextTick to get out of stack overflows.
            cb();
          }
        }, function () {
          // the rendering is finished; we should push out the last of the buffer.
          _this3.push(_this3.buffer);
          _this3.push(null);
        });
      } else {
        // continue with the rendering.
        this.continuation();
      }
    }
  }]);

  return RenderStream;
})(stream.Readable);

function renderToStringStream(element, res) {
  !ReactElement.isValidElement(element) ? process.env.NODE_ENV !== 'production' ? invariant(false, 'renderToStringStream(): You must pass a valid ReactElement.') : invariant(false) : undefined;

  var transaction;

  // NOTE that we never change this, which means that client rendering code cannot be used
  // in conjunction with this code.
  ReactUpdates.injection.injectBatchingStrategy(ReactServerBatchingStrategy);

  var id = ReactInstanceHandles.createReactRootID();
  transaction = ReactServerRenderingTransaction.getPooled(false);

  if (res) {
    console.warn("You are using version 0.1.x of the API, which has been deprecated.");
    var hash = rollingAdler32("");
    var readable = transaction.perform(function () {
      var buffer = "";
      var componentInstance = instantiateReactComponent(element, null);
      componentInstance.mountComponentAsync(id, transaction, emptyObject, function (text, cb) {
        hash = rollingAdler32(text, hash);
        buffer += text;
        if (buffer.length >= 16 * 1024) {
          res.write(buffer);
          buffer = "";
          process.nextTick(cb);
        } else {
          cb();
        }
      }, function () {
        res.write(buffer);
        res.end();
      });
      return null;
    }, null);
    return new Promise(function (resolve, reject) {
      res.on("end", function () {
        resolve(hash.hash());
      });
    });
  } else {
    var readable = transaction.perform(function () {
      var componentInstance = instantiateReactComponent(element, null);
      return new RenderStream(componentInstance, id, transaction, emptyObject);
    }, null);

    readable.on("end", function () {
      ReactServerRenderingTransaction.release(transaction);
      // Revert to the DOM batching strategy since these two renderers
      // currently share these stateful modules.
      ReactUpdates.injection.injectBatchingStrategy(ReactDefaultBatchingStrategy);
    });

    // since Adler32Stream has a .hash property, this automagically adds that property to the result.
    return readable.pipe(new Adler32Stream());
  }
}

/**
 * @param {ReactElement} element
 * @return {string} the HTML markup, without the extra React ID and checksum
 * (for generating static pages)
 */
function renderToStaticMarkupStream(element) {
  !ReactElement.isValidElement(element) ? process.env.NODE_ENV !== 'production' ? invariant(false, 'renderToStaticMarkupStream(): You must pass a valid ReactElement.') : invariant(false) : undefined;

  var transaction;

  // NOTE that we never change this, which means that client rendering code cannot be used
  // in conjunction with this code.
  ReactUpdates.injection.injectBatchingStrategy(ReactServerBatchingStrategy);

  var id = ReactInstanceHandles.createReactRootID();
  transaction = ReactServerRenderingTransaction.getPooled(true);

  var readable = transaction.perform(function () {
    var componentInstance = instantiateReactComponent(element, null);
    return new RenderStream(componentInstance, id, transaction, emptyObject);
  }, null);

  readable.on("end", function () {
    ReactServerRenderingTransaction.release(transaction);
  });

  return readable;
}

module.exports = {
  renderToStringStream: renderToStringStream,
  renderToStaticMarkupStream: renderToStaticMarkupStream
};